@@ -33,9 +33,10 @@ import collections
 import threading
 import multiprocessing as mp
 import queue
+from datetime import datetime
+from enum import Enum
 
 from ipaddress import ip_address
-from zope.interface import Interface
 
 ###
 ### Utility Classes
@@ -90,10 +91,19 @@ class SemaphoreN(threading.BoundedSemaphore):
         while self.acquire(blocking=False):
             pass
 
+class Conn(Enum):
+    OK = 0
+    FAILED = 1
+    TIMEOUT = 2
+    SKIPPED = 3
+
+Connection = collections.namedtuple("Connection", ["client", "port", "state", "tstart"])
+
 QUEUE_SIZE = 1000
 QUEUE_SLEEP = 0.5
 
-SHUTDOWN_SENTINEL = None
+SHUTDOWN_SENTINEL = "SHUTDOWN_SENTINEL"
+NO_RESULT = None
 NO_FLOW = None
 
 class Spider:
@@ -106,46 +116,32 @@ class Spider:
 
     """
 
-    def __init__(self):
+    def __init__(self, worker_count, libtrace_uri, args):
         """
-        Bare minimum initalisation for a pathspider plugin.
-
-        .. warning::
-         This function should not be overloaded by any plugin. Its purpose here
-         is only to set the "activated" flag to false, to prevent the plugin
-         functions being used before it has been activated.
-        """
-
-        self.activated = False
-
-    def activate(self, worker_count, libtrace_uri):
-        """
-        The activate function performs initialisation of a pathspider plugin.
+        The initialisation of a pathspider plugin.
 
         :param worker_count: The number of workers to use.
         :type worker_count: int
         :param libtrace_uri: The URI to pass to the Observer to describe the
                              interface on which packets should be captured.
         :type libtrace_uri: str
-        :see also: :func:`pathspider.base.ISpider.activate() <ISpider.activate>`
 
         It is expected that this function will be overloaded by plugins, though
-        the plugin should always make a call to the activate() function of the
+        the plugin should always make a call to the __init__() function of the
         abstract Spider class as this initialises all of the base functionality:
 
         .. code-block:: python
 
-         super().activate(worker_count=worker_count,
+         super().__init__(worker_count=worker_count,
                           libtrace_uri=libtrace_uri,
-                          check_interrupt=check_interrupt)
+                          args=args)
 
         This can be used to initialise any variables which may be required in
-        the object. Do not initialise any variables in the __init__ method, or
-        perform any other operations there as all plugins must be instantiated
-        in order to be loaded and this will cause unnecessary delays in the
-        starting of pathspider.
+        the object.
         """
 
+        self.args = args
+
         self.activated = True
         self.running = False
         self.stopping = False
@@ -156,16 +152,6 @@ class Spider:
         self.active_worker_lock = threading.Lock()
 
         self.libtrace_uri = libtrace_uri
-#        self.check_interrupt = check_interrupt
-
-        self.sem_config_zero = SemaphoreN(worker_count)
-        self.sem_config_zero.empty()
-        self.sem_config_zero_rdy = SemaphoreN(worker_count)
-        self.sem_config_zero_rdy.empty()
-        self.sem_config_one = SemaphoreN(worker_count)
-        self.sem_config_one.empty()
-        self.sem_config_one_rdy = SemaphoreN(worker_count)
-        self.sem_config_one_rdy.empty()
 
         self.jobqueue = queue.Queue(QUEUE_SIZE)
         self.resqueue = queue.Queue(QUEUE_SIZE)
@@ -175,6 +161,8 @@ class Spider:
 
         self.restab = {}
         self.flowtab = {}
+        self.flowreap = collections.deque()
+        self.flowreap_size = min(self.worker_count * 100, 10000)
 
         self.outqueue = queue.Queue(QUEUE_SIZE)
 
@@ -182,40 +170,16 @@ class Spider:
 
         self.worker_threads = []
         self.configurator_thread = None
-#        self.interrupter_thread = None
         self.merger_thread = None
 
         self.observer_process = None
 
-#        self._worker_state = [ "not_started" ] * self.worker_count
+        # self._worker_state = [ "not_started" ] * self.worker_count
 
         self.lock = threading.Lock()
         self.exception = None
 
-    def configurator(self):
-        """
-        Thread which synchronizes on a set of semaphores and alternates
-        between two system states.
-        """
-        logger = logging.getLogger('pathspider')
-
-        while self.running:
-            logger.debug("setting config zero")
-            self.config_zero()
-            logger.debug("config zero active")
-            self.sem_config_zero.release_n(self.worker_count)
-            self.sem_config_one_rdy.acquire_n(self.worker_count)
-            logger.debug("setting config one")
-            self.config_one()
-            logger.debug("config one active")
-            self.sem_config_one.release_n(self.worker_count)
-            self.sem_config_zero_rdy.acquire_n(self.worker_count)
-
-        # In case the master exits the run loop before all workers have,
-        # these tokens will allow all workers to run through again,
-        # until the next check at the start of the loop
-        self.sem_config_zero.release_n(self.worker_count)
-        self.sem_config_one.release_n(self.worker_count)
+        self.conn_timeout = None
 
     def config_zero(self):
         """
@@ -223,7 +187,7 @@ class Spider:
         baseline measurements.
         """
 
-        raise NotImplementedError("Cannot instantiate an abstract Pathspider")
+        raise NotImplementedError("Cannot instantiate an abstract Spider")
 
     def config_one(self):
         """
@@ -231,130 +195,13 @@ class Spider:
         experimental measurements.
         """
 
-        raise NotImplementedError("Cannot instantiate an abstract Pathspider")
+        raise NotImplementedError("Cannot instantiate an abstract Spider")
 
-    # def interrupter(self):
-    #     if self.check_interrupt is None:
-    #         return
-
-    #     logger = logging.getLogger('pathspider')
-    #     while self.running:
-    #         if self.check_interrupt():
-    #             logger.warning("spider interrupted")
-    #             logger.warning("trying to abort %d jobs", self.jobqueue.qsize())
-    #             while not self.jobqueue.empty():
-    #                 self.jobqueue.get()
-    #                 self.jobqueue.task_done()
-    #             self.stop()
-    #             break
-    #         time.sleep(5)
-
-    def worker(self, worker_number):
-        """
-        This function provides the logic for the worker threads.
-
-        :param worker_number: The unique number of the worker.
-        :type worker_number: int
-
-        The workers operate as continuous loops:
-
-         * Fetch next job from the job queue
-         * Perform pre-connection operations
-         * Acquire a lock for "config_zero"
-         * Perform the "config_zero" connection
-         * Release "config_zero"
-         * Acquire a lock for "config_one"
-         * Perform the "config_one" connection
-         * Release "config_one"
-         * Perform post-connection operations for config_zero and pass the
-           result to the merger
-         * Perform post-connection operations for config_one and pass the
-           result to the merger
-         * Do it all again
-        
-        If the job fetched is the SHUTDOWN_SENTINEL, then the worker will
-        terminate as this indicates that all the jobs have now been processed.
-        """
-
-        logger = logging.getLogger('pathspider')
-        worker_active = True
-
-        while self.running:
-            if worker_active:
-                try:
-                    job = self.jobqueue.get_nowait()
-
-                    # Break on shutdown sentinel
-                    if job == SHUTDOWN_SENTINEL:
-                        self.jobqueue.task_done()
-                        logger.debug("shutting down worker "+str(worker_number)+" on sentinel")
-                        #self._worker_state[worker_number] = "shutdown_sentinel"
-                        worker_active = False
-                        with self.active_worker_lock:
-                            self.active_worker_count -= 1
-                            logger.debug(str(self.active_worker_count)+" workers still active")
-                        continue
-
-                    logger.debug("got a job: "+repr(job))
-                except queue.Empty:
-                    #logger.debug("no job available, sleeping")
-                    # spin the semaphores
-                    self.sem_config_zero.acquire()
-                    #self._worker_state[worker_number] = "sleep_0"
-                    time.sleep(QUEUE_SLEEP)
-                    self.sem_config_one_rdy.release()
-                    self.sem_config_one.acquire()
-                    #self._worker_state[worker_number] = "sleep_1"
-                    time.sleep(QUEUE_SLEEP)
-                    self.sem_config_zero_rdy.release()
-                else:
-                    # Hook for preconnection
-                    #self._worker_state[worker_number] = "preconn"
-                    pcs = self.pre_connect(job)
-
-                    # Wait for configuration zero
-                    #self._worker_state[worker_number] = "wait_0"
-                    self.sem_config_zero.acquire()
-
-                    # Connect in configuration zero
-                    #self._worker_state[worker_number] = "conn_0"
-                    conn0 = self.connect(job, pcs, 0)
-
-                    # Wait for configuration one
-                    #self._worker_state[worker_number] = "wait_1"
-                    self.sem_config_one_rdy.release()
-                    self.sem_config_one.acquire()
-
-                    # Connect in configuration one
-                    #self._worker_state[worker_number] = "conn_1"
-                    conn1 = self.connect(job, pcs, 1)
-
-                    # Signal okay to go to configuration zero
-                    self.sem_config_zero_rdy.release()
-
-                    # Pass results on for merge
-                    #self._worker_state[worker_number] = "postconn_0"
-                    self.resqueue.put(self.post_connect(job, conn0, pcs, 0))
-                    #self._worker_state[worker_number] = "postconn_1"
-                    self.resqueue.put(self.post_connect(job, conn1, pcs, 1))
-
-                    #self._worker_state[worker_number] = "done"
-                    logger.debug("job complete: "+repr(job))
-                    self.jobqueue.task_done()
-            else: # not worker_active, spin the semaphores
-                self.sem_config_zero.acquire()
-                #self._worker_state[worker_number] = "shutdown_0"
-                time.sleep(QUEUE_SLEEP)
-                with self.active_worker_lock:
-                    if self.active_worker_count <= 0:
-                        #self._worker_state[worker_number] = "shutdown_complete"
-                        break
-                self.sem_config_one_rdy.release()
-                self.sem_config_one.acquire()
-                #self._worker_state[worker_number] = "shutdown_1"
-                time.sleep(QUEUE_SLEEP)
-                self.sem_config_zero_rdy.release()
+    def configurator(self):
+        raise NotImplementedError("Cannot instantiate an abstract Spider")
 
+    def worker(self):
+        raise NotImplementedError("Cannot instantiate an abstract Spider")
 
     def pre_connect(self, job):
         """
@@ -437,7 +284,7 @@ class Spider:
         This function is called by the base Spider logic to get an instance
         of :class:`pathspider.observer.Observer` configured with the function
         chains that are requried by the plugin.
-        
+
         This method is not implemented in the abstract
         :class:`pathspider.base.Spider` class and must be implemented by any
         plugin.
@@ -457,8 +304,16 @@ class Spider:
         merging_flows = True
         merging_results = True
 
-        while self.running and merging_results:
-            if self.flowqueue.qsize() >= self.resqueue.qsize():
+        # merge_cycles = 0
+
+        while self.running and (merging_results or merging_flows):
+
+            # if merge_cycles % 20 == 0:
+            #     for wn in range(0, self.worker_count):
+            #         logger.debug("worker %3u: %s" % (wn, self._worker_state[wn]))
+            # merge_cycles += 1
+
+            if merging_flows and self.flowqueue.qsize() >= self.resqueue.qsize():
                 try:
                     flow = self.flowqueue.get_nowait()
                 except queue.Empty:
@@ -480,19 +335,27 @@ class Spider:
                     elif flowkey in self.flowtab:
                         logger.debug("won't merge duplicate flow")
                     else:
-                        # FIXME: How to keep flowtab from 
-                        # exploding with unrelated flows?
-                        # We need a timer queue for flow expiry. 
-                        # See Issue #30
+                        # Create a new flow
                         self.flowtab[flowkey] = flow
 
-            else:
+                        # And reap the oldest, if the reap queue is full
+                        self.flowreap.append(flowkey)
+                        if len(self.flowreap) > self.flowreap_size:
+                            try:
+                                del self.flowtab[self.flowreap.popleft()]
+                            except KeyError:
+                                pass
+
+            elif merging_results:
                 try:
                     res = self.resqueue.get_nowait()
                 except queue.Empty:
                     time.sleep(QUEUE_SLEEP)
                     logger.debug("result queue is empty")
                 else:
+                    if res == NO_RESULT:
+                        # handle skipped results
+                        continue
                     if res == SHUTDOWN_SENTINEL:
                         merging_results = False
                         logger.debug("stopping result merging on sentinel")
@@ -513,10 +376,10 @@ class Spider:
 
                     self.resqueue.task_done()
 
-        # Both shutdown markers received. 
-        # Call merge on all remaining entries in the results table 
+        # Both shutdown markers received.
+        # Call merge on all remaining entries in the results table
         # with null flows.
-        # Commented out for now; see https://github.com/mami-project/pathspider/issues/29 
+        # Commented out for now; see https://github.com/mami-project/pathspider/issues/29
         for res_item in self.restab.items():
             res = res_item[1]
             self.merge(NO_FLOW, res)
@@ -589,7 +452,7 @@ class Spider:
             self.observer = self.create_observer()
             self.observer_process = mp.Process(
                 args=(self.observer.run_flow_enqueuer,
-                      self.flowqueue, 
+                      self.flowqueue,
                       self.observer_shutdown_queue),
                 target=self.exception_wrapper,
                 name='observer',
@@ -633,41 +496,31 @@ class Spider:
                 worker_thread.start()
             logger.debug("workers up")
 
-            # if self.check_interrupt is not None:
-            #     self.interrupter_thread = threading.Thread(
-            #         args=(self.interrupter,),
-            #         target=self.exception_wrapper,
-            #         name="interrupter",
-            #         daemon=True)
-            #     self.interrupter_thread.start()
-            #     logger.debug("interrupter up")
-
     def shutdown(self):
         """
-        Shut down PathSpider in an orderly fashion, 
-        ensuring that all queued jobs complete, 
+        Shut down PathSpider in an orderly fashion,
+        ensuring that all queued jobs complete,
         and all available results are merged.
 
         """
         logger = logging.getLogger('pathspider')
 
-        logger.info("shutting down pathspider")
-        
+        logger.info("beginning shutdown")
+
         with self.lock:
             # Set stopping flag
             self.stopping = True
 
-            # Place two shutdown sentinels per worker
-            # in the job queue FIXME HACK
-            for i in range(self.worker_count):
-                self.jobqueue.put(SHUTDOWN_SENTINEL) 
+            # Put a bunch of shutdown signals in the job queue
+            for i in range(self.worker_count * 2):
+                self.jobqueue.put(SHUTDOWN_SENTINEL)
 
             # Wait for worker threads to shut down
             for worker in self.worker_threads:
                 if threading.current_thread() != worker:
                     logger.debug("joining worker: " + repr(worker))
                     worker.join()
-            logger.debug("all workers joined")            
+            logger.debug("all workers joined")
 
             # Tell observer to shut down
             self.observer_shutdown_queue.put(True)
@@ -692,7 +545,7 @@ class Spider:
             # Join configurator
             # if threading.current_thread() != self.configurator_thread:
             #     self.configurator_thread.join()
-        
+
             self.stopping = False
 
         logger.info("shutdown complete")
@@ -737,22 +590,23 @@ class Spider:
             if threading.current_thread() != worker:
                 logger.debug("joining worker: " + repr(worker))
                 worker.join()
-        logger.debug("all workers joined")           
+        logger.debug("all workers joined")
 
-        if threading.current_thread() != self.configurator_thread:
+        if self.configurator_thread and \
+                (threading.current_thread() != self.configurator_thread):
             self.configurator_thread.join()
-        logger.debug("configurator joined")           
-        
+            logger.debug("configurator joined")
+
         if threading.current_thread() != self.merger_thread:
-            self.merger_thread.join() 
-        logger.debug("merger joined")           
+            self.merger_thread.join()
+            logger.debug("merger joined")
 
         self.observer_process.join()
         logger.debug("observer joined")
 
         self.outqueue.put(SHUTDOWN_SENTINEL)
         logger.info("termination complete")
-           
+
     def add_job(self, job):
         """
         Adds a job to the job queue.
@@ -766,32 +620,187 @@ class Spider:
 
         self.jobqueue.put(job)
 
-# def local_address(ipv=4, target="path-ams.corvid.ch", port=53):
-#     if ipv == 4:
-#         addrfamily = socket.AF_INET
-#     elif ipv == 6:
-#         addrfamily = socket.AF_INET6
-#     else:
-#         assert False
 
-#     try:
-#         sock = socket.socket(addrfamily, socket.SOCK_DGRAM)
-#         sock.connect((target, port))
-#         return ip_address(sock.getsockname()[0])
-#     except:
-#         #FIXME: What exceptions do we expect?
-#         return None
+class SynchronizedSpider(Spider):
 
-class ISpider(Interface): # pylint: disable=E0239
-    """
-    The ISpider interface defines the expected interface for PATHspider plugins.
-    """
+    def __init__(self, worker_count, libtrace_uri, args):
+        super().__init__(worker_count, libtrace_uri, args)
 
-    def activate(self, worker_count, libtrace_uri):
-        pass
+        # create semaphores for synchronizing configurations
+        self.sem_config_zero = SemaphoreN(worker_count)
+        self.sem_config_zero.empty()
+        self.sem_config_zero_rdy = SemaphoreN(worker_count)
+        self.sem_config_zero_rdy.empty()
+        self.sem_config_one = SemaphoreN(worker_count)
+        self.sem_config_one.empty()
+        self.sem_config_one_rdy = SemaphoreN(worker_count)
+        self.sem_config_one_rdy.empty()
 
     def configurator(self):
-        pass
+        """
+        Thread which synchronizes on a set of semaphores and alternates
+        between two system states.
+        """
+        logger = logging.getLogger('pathspider')
+
+        while self.running:
+            logger.debug("setting config zero")
+            self.config_zero()
+            logger.debug("config zero active")
+            self.sem_config_zero.release_n(self.worker_count)
+            self.sem_config_one_rdy.acquire_n(self.worker_count)
+            logger.debug("setting config one")
+            self.config_one()
+            logger.debug("config one active")
+            self.sem_config_one.release_n(self.worker_count)
+            self.sem_config_zero_rdy.acquire_n(self.worker_count)
+
+        # In case the master exits the run loop before all workers have,
+        # these tokens will allow all workers to run through again,
+        # until the next check at the start of the loop
+        self.sem_config_zero.release_n(self.worker_count)
+        self.sem_config_one.release_n(self.worker_count)
+
+    def worker(self, worker_number):
+        """
+        This function provides the logic for
+        configuration-synchronized worker threads.
+
+        :param worker_number: The unique number of the worker.
+        :type worker_number: int
+
+        The workers operate as continuous loops:
+
+         * Fetch next job from the job queue
+         * Perform pre-connection operations
+         * Acquire a lock for "config_zero"
+         * Perform the "config_zero" connection
+         * Release "config_zero"
+         * Acquire a lock for "config_one"
+         * Perform the "config_one" connection
+         * Release "config_one"
+         * Perform post-connection operations for config_zero and pass the
+           result to the merger
+         * Perform post-connection operations for config_one and pass the
+           result to the merger
+         * Do it all again
+
+        If the job fetched is the SHUTDOWN_SENTINEL, then the worker will
+        terminate as this indicates that all the jobs have now been processed.
+        """
+
+        logger = logging.getLogger('pathspider')
+        worker_active = True
+
+        while self.running:
+            if worker_active:
+                try:
+                    job = self.jobqueue.get_nowait()
+
+                    # Break on shutdown sentinel
+                    if job == SHUTDOWN_SENTINEL:
+                        self.jobqueue.task_done()
+                        logger.debug("shutting down worker "+str(worker_number)+" on sentinel")
+                        # self._worker_state[worker_number] = "shutdown_sentinel"
+                        worker_active = False
+                        with self.active_worker_lock:
+                            self.active_worker_count -= 1
+                            logger.debug(str(self.active_worker_count)+" workers still active")
+                        continue
+
+                    logger.debug("got a job: "+repr(job))
+                except queue.Empty:
+                    #logger.debug("no job available, sleeping")
+                    # spin the semaphores
+                    self.sem_config_zero.acquire()
+                    # self._worker_state[worker_number] = "sleep_0"
+                    time.sleep(QUEUE_SLEEP)
+                    self.sem_config_one_rdy.release()
+                    self.sem_config_one.acquire()
+                    # self._worker_state[worker_number] = "sleep_1"
+                    time.sleep(QUEUE_SLEEP)
+                    self.sem_config_zero_rdy.release()
+                else:
+                    # Hook for preconnection
+                    # self._worker_state[worker_number] = "preconn"
+                    pcs = self.pre_connect(job)
+
+                    # Wait for configuration zero
+                    # self._worker_state[worker_number] = "wait_0"
+                    self.sem_config_zero.acquire()
+
+                    # Connect in configuration zero
+                    # self._worker_state[worker_number] = "conn_0"
+                    conn0 = self.connect(job, pcs, 0)
+
+                    # Wait for configuration one
+                    # self._worker_state[worker_number] = "wait_1"
+                    self.sem_config_one_rdy.release()
+                    self.sem_config_one.acquire()
+
+                    # Connect in configuration one
+                    # self._worker_state[worker_number] = "conn_1"
+                    conn1 = self.connect(job, pcs, 1)
+
+                    # Signal okay to go to configuration zero
+                    self.sem_config_zero_rdy.release()
+
+                    # Pass results on for merge
+                    # self._worker_state[worker_number] = "postconn_0"
+                    self.resqueue.put(self.post_connect(job, conn0, pcs, 0))
+                    # self._worker_state[worker_number] = "postconn_1"
+                    self.resqueue.put(self.post_connect(job, conn1, pcs, 1))
+
+                    # self._worker_state[worker_number] = "done"
+                    logger.debug("job complete: "+repr(job))
+                    self.jobqueue.task_done()
+            else: # not worker_active, spin the semaphores
+                self.sem_config_zero.acquire()
+                # self._worker_state[worker_number] = "shutdown_0"
+                time.sleep(QUEUE_SLEEP)
+                with self.active_worker_lock:
+                    if self.active_worker_count <= 0:
+                        # self._worker_state[worker_number] = "shutdown_complete"
+                        break
+                self.sem_config_one_rdy.release()
+                self.sem_config_one.acquire()
+                # self._worker_state[worker_number] = "shutdown_1"
+                time.sleep(QUEUE_SLEEP)
+                self.sem_config_zero_rdy.release()
+
+    def tcp_connect(self, job):
+        """
+        This helper function will perform a TCP connection. It will not perform
+        any special action in the event that this is the experimental flow,
+        it only performs a TCP connection. This function expects that
+        self.conn_timeout has been set to a sensible value.
+        """
+
+        if self.conn_timeout is None:
+            raise RuntimeError("Plugin did not set TCP connect timeout.")
+
+        tstart = str(datetime.utcnow())
+
+        if ":" in job[0]:
+            sock = socket.socket(socket.AF_INET6)
+        else:
+            sock = socket.socket(socket.AF_INET)
+
+        try:
+            sock.settimeout(self.conn_timeout)
+            sock.connect((job[0], job[1]))
+
+            return Connection(sock, sock.getsockname()[1], Conn.OK, tstart)
+        except TimeoutError:
+            return Connection(sock, sock.getsockname()[1], Conn.TIMEOUT, tstart)
+        except OSError:
+            return Connection(sock, sock.getsockname()[1], Conn.FAILED, tstart)
+
+
+class DesynchronizedSpider(Spider):
+
+    def __init__(self, worker_count, libtrace_uri, args):
+        super().__init__(worker_count, libtrace_uri, args)
 
     def config_zero(self):
         pass
@@ -799,39 +808,93 @@ class ISpider(Interface): # pylint: disable=E0239
     def config_one(self):
         pass
 
+    def configurator(self):
+        """
+        Since there is no need for a configurator thread in a
+        desynchronized spider, this thread is a no-op
+        """
+        logger = logging.getLogger('pathspider')
+        logger.info("configurations are not synchronized")
+
     def worker(self, worker_number):
-        pass
+        """
+        This function provides the logic for
+        configuration-synchronized worker threads.
 
-    def pre_connect(self, job):
-        pass
+        :param worker_number: The unique number of the worker.
+        :type worker_number: int
 
-    def connect(self, job, pcs, config):
-        pass
+        The workers operate as continuous loops:
 
-    def post_connect(self, job, conn, pcs, config):
-        pass
+         * Fetch next job from the job queue
+         * Perform pre-connection operations
+         * Acquire a lock for "config_zero"
+         * Perform the "config_zero" connection
+         * Release "config_zero"
+         * Acquire a lock for "config_one"
+         * Perform the "config_one" connection
+         * Release "config_one"
+         * Perform post-connection operations for config_zero and pass the
+           result to the merger
+         * Perform post-connection operations for config_one and pass the
+           result to the merger
+         * Do it all again
 
-    def create_observer(self):
-        pass
+        If the job fetched is the SHUTDOWN_SENTINEL, then the worker will
+        terminate as this indicates that all the jobs have now been processed.
+        """
 
-    def merger(self):
-        pass
+        logger = logging.getLogger('pathspider')
+        worker_active = True
 
-    def merge(self, flow, res):
-        pass
+        while self.running:
+            if worker_active:
+                try:
+                    job = self.jobqueue.get_nowait()
 
-    def exception_wrapper(self, target, *args, **kwargs):
-        pass
+                    # Break on shutdown sentinel
+                    if job == SHUTDOWN_SENTINEL:
+                        self.jobqueue.task_done()
+                        logger.debug("shutting down worker "+str(worker_number)+" on sentinel")
+                        # self._worker_state[worker_number] = "shutdown_sentinel"
+                        worker_active = False
+                        with self.active_worker_lock:
+                            self.active_worker_count -= 1
+                            logger.debug(str(self.active_worker_count)+" workers still active")
+                        continue
 
-    def start(self):
-        pass
+                    logger.debug("got a job: "+repr(job))
+                except queue.Empty:
+                    time.sleep(QUEUE_SLEEP)
+                else:
+                    # Hook for preconnection
+                    # self._worker_state[worker_number] = "preconn"
+                    pcs = self.pre_connect(job)
 
-    def shutdown(self):
-        pass
+                    # Connect in configuration zero
+                    # self._worker_state[worker_number] = "conn_0"
+                    conn0 = self.connect(job, pcs, 0)
 
-    def terminate(self):
-        pass
+                    # Connect in configuration one
+                    # self._worker_state[worker_number] = "conn_1"
+                    conn1 = self.connect(job, pcs, 1)
 
-    def add_job(self, job):
-        pass
+                    # Pass results on for merge
+                    # self._worker_state[worker_number] = "postconn_0"
+                    self.resqueue.put(self.post_connect(job, conn0, pcs, 0))
+                    # self._worker_state[worker_number] = "postconn_1"
+                    self.resqueue.put(self.post_connect(job, conn1, pcs, 1))
 
+                    # self._worker_state[worker_number] = "done"
+                    logger.debug("job complete: "+repr(job))
+                    self.jobqueue.task_done()
+            elif not self.stopping:
+                time.sleep(QUEUE_SLEEP)
+            else:
+                # self._worker_state[worker_number] = "shutdown_complete"
+                break
+
+class PluggableSpider:
+    @staticmethod
+    def register_args(subparsers):
+        raise NotImplementedError("Cannot register an abstract plugin")
